/* window.vala
 *
 * Copyright 2019 Vivien Kraus
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

[CCode (cname="VERSION")]
extern const string VERSION;

[CCode (cname="EXTERNAL_FILE_MANAGER_PROGRAM")]
extern const string EXTERNAL_FILE_MANAGER_PROGRAM;

namespace Tarot {

	[DBus (name = "org.freedesktop.FileManager1")]
	public interface FileManagerInterface: Object {
		public abstract void ShowFolders (string[] URIs, string StartupId) throws GLib.Error;
	}

	private class EventQueue {

		// If there are more than that number of events, an assert is
		// raised!
		private Tarot.GameEvent m_data[256];
		private size_t m_head;
		private size_t m_n;
		private bool timer_running = false;
		private bool deleted = false;
		private double next_delay = 0;

		public signal void new_event (Tarot.GameEvent e);
		public signal void catchup ();

		private void start_timer () {
			timer_running = true;
			double seconds = next_delay;
			next_delay = 1;
			switch (m_data[m_head].type ()) {
			case Tarot.GameEventT.SETUP_EVENT:
			case Tarot.GameEventT.DEAL_EVENT:
				next_delay = 0;
				break;
			case Tarot.GameEventT.DECL_EVENT:
				next_delay = 0.5;
				break;
			case Tarot.GameEventT.DOG_EVENT:
				next_delay = 3;
				break;
			default:
				next_delay = 1;
				break;
			}
			GLib.Timeout.add ((uint) (1000 * seconds), () => {
					timer_running = false;
					if (!deleted) {
						dequeue ();
					}
					return false;
				});
		}

		private void dequeue () {
			assert (m_n > 0);
			new_event (m_data[m_head]);
			m_head ++;
			m_n --;
			m_head %= m_data.length;
			if (m_n > 0) {
				assert (!timer_running);
				start_timer ();
			} else {
				catchup ();
			}
		}

		public void enqueue (Tarot.GameEvent e) {
			assert (m_n < m_data.length);
			size_t i = (m_head + m_n) % m_data.length;
			m_data[i] = e;
			if (m_n++ == 0) {
				assert (!timer_running);
				start_timer ();
			}
		}

		public void close () {
			deleted = true;
		}
	}

	[GtkTemplate (ui = "/eu/planete_kraus/Tarot/window.ui")]
	public class Window : Gtk.ApplicationWindow {

		[GtkChild]
		private Gtk.Stack pages;

		[GtkChild]
		private Tarot.Widget tarot_widget;

		[GtkChild]
		private Gtk.Box page_select_players;

		[GtkChild]
		private Gtk.Switch variant_call;

		[GtkChild]
		private Gtk.Button button_new_game;

		[GtkChild]
		private Gtk.PopoverMenu menu_popover;

		[GtkChild]
		private Gtk.LinkButton button_browse_games;

		[GtkCallback]
		private void show_new_game_page () {
			my_game = new Tarot.Game.alloc ();
			m_queue.close ();
			m_queue = new EventQueue ();
			m_queue.new_event.connect (add_server_event);
			m_queue.catchup.connect (() => { check_if_next (); });
			pages.visible_child = page_select_players;
			button_new_game.hide ();
		}

		[GtkCallback]
		private void new_game_3p () {
			start_new_game (3, false);
		}

		[GtkCallback]
		private void new_game_4p () {
			start_new_game (4, false);
		}

		[GtkCallback]
		private void new_game_5p () {
			start_new_game (5, variant_call.active);
		}

		[GtkCallback]
		private void open_info_dialog () {
			string[] artists = { "Individu Masqué", null };
			string[] authors = { "Vivien Kraus", null };
			var comments = _ ("If you find that the AI is too weak, send me your game files at vivien@planete-kraus.eu :)");
			var copyright = _ ("Copyright © 2019 Vivien Kraus");
			string[] documenters = { "Guess Who", null };
			var license_type = Gtk.License.GPL_3_0;
			var program_name = "eu.planete_kraus.Tarot";
			var version = VERSION;
			var website = "https://play-tarot.frama.io/tarot";
			var logo_icon_name = "eu.planete_kraus.Tarot";
			Gtk.show_about_dialog (this,
								   "artists", artists,
								   "authors", authors,
								   "comments", comments,
								   "copyright", copyright,
								   "documenters", documenters,
								   "license-type", license_type,
								   "program-name", program_name,
								   "version", version,
								   "website", website,
								   "logo-icon-name", logo_icon_name,
								   null);
			menu_popover.hide ();
		}

		[GtkCallback]
		private void show_menu_popover () {
			menu_popover.show_all ();
		}

		private Tarot.Game my_game;
		private Tarot.ServerGame server;
		private Tarot.StackingGame?[] ais;
		private Tarot.Player myself = 0;
		private EventQueue m_queue;

		private void add_server_event (Tarot.GameEvent e) {
			if (my_game.add_event (e) != Tarot.GameError.OK) {
				// The server is desynchronized...  This is impossible
				// for now that we are in the same program.
				assert (false);
			}
			tarot_widget.game = my_game;
		}

		private void check_if_next () {
			tarot_widget.game = my_game;
			tarot_widget.dont_ask ();
			Tarot.Player next;
			Tarot.Bid mini;
			if (my_game.get_hint_bid (out next, out mini) == Tarot.GameError.OK
				&& next == myself) {
				stderr.printf ("Asking bid >= %s\n".printf (mini.to_string ()));
				tarot_widget.ask_bid (mini);
			}
			bool allowed;
			if (my_game.get_hint_decl (out next, out allowed) == Tarot.GameError.OK
				&& next == myself) {
				stderr.printf ("Asking declaration <= %s\n".printf (allowed.to_string ()));
				tarot_widget.ask_declaration (allowed);
			}
			Tarot.Number mini_call;
			if (my_game.get_hint_call (out next, out mini_call) == Tarot.GameError.OK
				&& next == myself) {
				stderr.printf ("Asking call <= %s\n".printf (mini_call.to_string ()));
				tarot_widget.ask_call (mini_call);
			}
			size_t n_discard;
			Tarot.Card[] prio;
			Tarot.Card[] additional;
			if (my_game.get_hint_discard_alloc (out next, out n_discard, out prio, out additional) == Tarot.GameError.OK
				&& next == myself) {
				stderr.printf ("Asking discard: %lu cards\n", n_discard);
				tarot_widget.ask_discard (n_discard, prio, additional);
			}
			size_t n_simple, n_double, n_triple;
			if (my_game.get_hint_handful_alloc (out next, out n_simple, out n_double, out n_triple, out prio, out additional) == Tarot.GameError.OK
				&& next == myself
				&& n_simple <= prio.length + additional.length) {
				stderr.printf ("Asking handful: %lu, %lu or %lu cards\n", n_simple, n_double, n_triple);
				tarot_widget.ask_handful (n_simple, n_double, n_triple, prio, additional);
			}
			Tarot.Card[] playable;
			if (my_game.get_hint_card_alloc (out next, out playable) == Tarot.GameError.OK
				&& next == myself) {
				stderr.printf ("Asking card");
				tarot_widget.ask_play (playable);
			}
			int[] score;
			if (my_game.get_scores_alloc (out score) == Tarot.GameError.OK) {
				var date_now = new GLib.DateTime.now_utc ().format ("%Y-%m-%d+%T");
				var myself = myself.to_string_alloc ();
				var filename = (GLib.Environment.get_variable ("SAVEDGAMESDIR") + "/" + date_now + "_" + myself + ".tarot");
				var contents = my_game.save_to_xml_alloc ();
				var file = GLib.File.new_for_path (filename);
				try {
					var output = new DataOutputStream (file.create (FileCreateFlags.REPLACE_DESTINATION));
					output.put_string (contents);
				} catch (Error e) {
					stderr.printf (_ ("Error: could not log the game: %s\n"), e.message);
				}
			}
		}

		construct {
			m_queue = new EventQueue ();
			tarot_widget.select_bid.connect ((bid) => {
					stderr.printf ("Selected bid %s\n", bid.to_string ());
					try {
						server.bid ((uint) bid, myself);
					} catch (ServerError err) {
						assert (false);
					}
				});
			tarot_widget.select_declaration.connect ((decl) => {
					stderr.printf ("Selected decl %s\n", decl.to_string ());
					try {
						server.declare (decl, myself);
					} catch (ServerError err) {
						assert (false);
					}
				});
			tarot_widget.select_call.connect ((card) => {
					stderr.printf ("Selected call %s\n", ((Tarot.Card) card).to_string_alloc ());
					try {
						server.call (card, myself);
					} catch (ServerError err) {
						assert (false);
					}
				});
			tarot_widget.select_discard.connect ((cards) => {
					stderr.printf ("Selected discard\n");
					try {
						server.discard (cards, myself);
					} catch (ServerError err) {
						assert (false);
					}
				});
			tarot_widget.select_handful.connect ((cards) => {
					stderr.printf ("Selected handful\n");
					try {
						server.handful (cards, myself);
					} catch (ServerError err) {
						assert (false);
					}
				});
			tarot_widget.select_card.connect ((card) => {
					stderr.printf ("Selected call %s\n", ((Tarot.Card) card).to_string_alloc ());
					try {
						server.play (card, myself);
					} catch (ServerError err) {
						assert (false);
					}
					tarot_widget.dont_ask ();
				});
		}

		private void start_new_game (size_t n_players, bool with_call) {
			try {
				server = new ServerGame (n_players, with_call);
			} catch (ServerError e) {
				stderr.printf (_ ("Error: could not set up the game: %s\n"), e.message);
				GLib.Process.exit (1);
			}
			myself = prng.int_range (0, (int32) n_players);
			ais = new Tarot.StackingGame?[n_players];
			for (size_t i = 0; i < n_players; i++) {
				ais[i] = null;
				if (i != myself) {
					ais[i] = new Tarot.StackingGame (new ServerSeat ((Tarot.Player) i, server));
				}
			}
			string seed = new GLib.DateTime.now_utc ().to_string ();
			while (!server.deal_random (seed)) {
				seed += ".";
			}
			pages.visible_child = tarot_widget;
			server.event_setup.connect ((destination, n_players, with_call) => {
					var e = new Tarot.GameEvent.alloc_setup (n_players, with_call);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
					// The MCTS have received the event through their seats.
				});
			server.event_deal.connect ((destination, cards) => {
					var e = new Tarot.GameEvent.alloc_deal (destination, (Tarot.Card[]) cards);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_bid.connect ((destination, bid) => {
					var e = new Tarot.GameEvent.alloc_bid ((Tarot.Bid) bid);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_decl.connect ((destination, decl) => {
					var e = new Tarot.GameEvent.alloc_decl (decl);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_call.connect ((destination, call) => {
					var e = new Tarot.GameEvent.alloc_call (call);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_dog.connect ((destination, cards) => {
					var e = new Tarot.GameEvent.alloc_dog ((Tarot.Card[]) cards);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_discard.connect ((destination, cards) => {
					var e = new Tarot.GameEvent.alloc_discard ((Tarot.Card[]) cards);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_handful.connect ((destination, cards) => {
					var e = new Tarot.GameEvent.alloc_handful ((Tarot.Card[]) cards);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_play.connect ((destination, card) => {
					var e = new Tarot.GameEvent.alloc_card (card);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.next.connect ((who) => {
					stderr.printf ("Next: %s\n", ((Tarot.Player) who).to_string_alloc ());
				});
			server.end.connect ((scores) => {
				});
			server.subscribe (myself);
			button_new_game.show ();
			tarot_widget.dont_ask ();
		}

		private GLib.Rand prng = new GLib.Rand ();

		public Window (Gtk.Application app) {
			Object (application: app);
			var screen = Gdk.Screen.get_default ();
			var provider = new Gtk.CssProvider ();
			provider.load_from_resource ("/eu/planete_kraus/Tarot/style.css");
			Gtk.StyleContext.add_provider_for_screen (screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION);
			button_browse_games.uri = "file://" + GLib.Environment.get_variable ("SAVEDGAMESDIR");
			show_new_game_page ();
		}
	}
}
