/*
 * tarot_cnn_features.c
 *
 * Copyright (C) 2019 Vivien Kraus <vivien@planete-kraus.eu>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "program_commons.h"
#include <tuneconfig.h>

static void parse_options (int argc, char *argv[]);
static void print_row (size_t n, const double *row);

int
main (int argc, char *argv[])
{
  TarotGame *game = NULL;
  double *data;
  size_t n;
  const TarotGameEvent *e;
  TarotGameIterator *it;
  TarotGame *rebuilt;
  tarot_program_setup ();
  parse_options (argc, argv);
  game = tarot_program_read_game ();
  rebuilt = tarot_game_alloc ();
  it = tarot_game_iterator (game);
  while ((e = tarot_game_iterator_next_value (it)) != NULL)
    {
      TarotCard *cards;
      TarotPlayer *players;
      size_t n_cards;
      size_t n_players;
      TarotPlayer next;
      int has_next = (tarot_game_get_next (rebuilt, &next) == TAROT_GAME_OK);
      if (has_next
	  && tarot_game_event_get_deal_all_alloc (e, &n_players, &players) == TAROT_EVENT_OK)
	{
	  TarotPlayer myself = next;
	  TarotCard my_cards[78];
	  size_t n_my_cards = 0;
	  TarotCard i;
	  TarotGameEvent *my_deal = NULL;
	  for (i = 0; i < n_players; i++)
	    {
	      if (players[i] == myself)
		{
		  my_cards[n_my_cards++] = i;
		}
	    }
	  my_deal = tarot_game_event_alloc_deal (myself, n_my_cards, my_cards);
	  if (tarot_game_add_event (rebuilt, my_deal) != 0)
	    {
	      assert (0);
	    }
	  tarot_game_event_free (my_deal);
	  free (players);
	}
      else if (has_next
	       && tarot_game_event_get_discard_alloc (e, &n_cards, &cards) == TAROT_EVENT_OK)
	{
	  TarotCard *full_discard = NULL;
	  size_t n_full;
	  TarotCard *partial_discard = NULL;
	  size_t n_partial;
	  TarotPlayer taker;
	  TarotGameEvent *my_discard = NULL;
	  tarot_game_get_taker (rebuilt, &taker);
	  if (tarot_game_get_full_discard_alloc (game, &n_full, &full_discard) != TAROT_GAME_OK)
	    {
	      full_discard = NULL;
	    }
	  if (tarot_game_get_public_discard_alloc (game, &n_partial, &partial_discard) != TAROT_GAME_OK)
	    {
	      partial_discard = NULL;
	    }
	  if (full_discard != NULL && partial_discard != NULL)
	    {
	      if (next == taker)
		{
		  my_discard = tarot_game_event_alloc_discard (n_full, full_discard);
		}
	      else
		{
		  my_discard = tarot_game_event_alloc_discard (n_partial, partial_discard);
		}
	      tarot_features_cnn_alloc (rebuilt, &data, &n);
	      print_row (n, data);
	      free (data);
	    }
	  else
	    {
	      my_discard = tarot_game_event_dup (e);
	    }
	  if (tarot_game_add_event (rebuilt, my_discard) != 0)
	    {
	      assert (0);
	    }
	  tarot_game_event_free (my_discard);
	  free (partial_discard);
	  free (full_discard);
	  free (cards);
	}
      else if (tarot_game_add_event (rebuilt, e) != 0)
	{
	  assert (0);
	}
      tarot_features_cnn_alloc (rebuilt, &data, &n);
      print_row (n, data);
      free (data);
    }
  return EXIT_SUCCESS;
}

static void
parse_options (int argc, char *argv[])
{
  int c;
  while (1)
    {
      static struct option long_options[] =
        {
	 {N_ ("help"), no_argument, 0, 'h'},
	 {N_ ("version"), no_argument, 0, 'v'},
	 {0, 0, 0, 0}
        };
      int option_index = 0;
      for (option_index = 0; long_options[option_index].name != NULL; option_index++)
	{
	  long_options[option_index].name = _ (long_options[option_index].name);
	}
      option_index = 0;
      c = getopt_long (argc, argv, "hv", long_options, &option_index);
      if (c == -1)
        break;
      switch (c)
        {
        case 0:
          break;

	case 'h':
	  printf (_ ("Usage: tarot-cnn-features [OPTIONS]... < GAME > FEATURES\n\
Extract the tensor-shaped game features.  See the info manual for the\n\
details.  Options:\n\
- -h, --help: print this message and exit;\n\
- -v, --version: print the program version and exit.\n\
"));
	  exit (EXIT_SUCCESS);
	  break;

	case 'v':
	  printf (_ ("%s (libtarot %s)\n\
\n\
Copyright © 2019 Vivien Kraus\nThis is free software; see the source for copying conditions.  There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
		  "tarot-cnn-features",
		  tarot_libtarot_version ());
	  exit (EXIT_SUCCESS);
	  break;

        case '?':
          /* getopt_long already printed an error message. */
	  exit (EXIT_FAILURE);
          break;

        default:
          abort ();
        }
    }
  if (optind < argc)
    {
      fprintf (stderr, _ ("Error: extra arguments:\n"));
      while (optind < argc)
        fprintf (stderr, _ ("- %s\n"), argv[optind++]);
      exit (EXIT_FAILURE);
    }
}

static void
print_row (size_t n, const double *data)
{
  size_t i;
  for (i = 0; i < n; i++)
    {
      if (i != 0)
	{
	  printf (",");
	}
      if (data[i] == 0.0)
	{
	  printf ("0");
	}
      else if (data[i] == 1.0)
	{
	  printf ("1");
	}
      else
	{
	  printf ("%f", data[i]);
	}
    }
  printf ("\n");
}
